var  express   = require('express')
    //,routes    = require('../routes')
    ,util          = require('util')
    ,redis         = require("redis")
    ,fs            = require('fs')
    ,_             = require('underscore')
    ,exec          = require('child_process').exec
    ,passport      = require('passport')
    ,path          = require('path')
    ,SESSIONS  = {}
    ,REDIS_CLIENTS = {};

require('./config/passport')(passport);

var createWebServer = function(config) {
    var self = this
        ,app = express.createServer()
        ,io = require('socket.io').listen(app)
        ,devicesSocket = io.of('/devices')
        ,ACTIVE_SESSIONS = {}
        ,dbc = redis.createClient(config.redisPort);
        this.socket = devicesSocket;

    if( !(process.env.NODE_ENV) ){
        app.set('env', 'production');
    }

    app.register('.html', {
        compile: function(str, options){
            try {
                var template = _.template(str);
                return function(locals){
                    return template(locals);
                };
            }
            catch (e) {
                console.log('compile html', e);
            }
        }
    });


    // Configuration
    app.set('view options', {
        layout: false
    });
    app.configure(function(){
        app.set('views', __dirname + '/views');
        //app.set('view engine', 'jade');
        app.set('view engine', 'html');
        app.use(express.methodOverride());
        app.use(express.bodyParser());
        app.use(express.cookieParser());
        app.use(express.session({ secret: 'nicol_marius_dovi_gabe_dennis_ilya_chris_roger_ryan_ilya_richard' }));
        app.use(passport.initialize());
        app.use(passport.session());
        app.use('/static', express.static(__dirname + '/static'));
        app.use(express.static(__dirname + '/public'));
        app.use(app.router);
        app.use(express.logger());
        app.use(express.errorHandler());

    });

    app.configure('development', function(){
        app.use(express.errorHandler({
            dumpExceptions: true,
            showStack: true
        }));
        io.set('log level', 10);
    });

    app.configure('production', function(){
        app.use(express.errorHandler());
        io.set('log level', 0);

    });
    try{
    require('./routes/auth.js')(app, passport);
    app.all('*', requireAuthentication);
    require('./routes/user.js')(app, checkAdmin);
    require('./routes/server_settings.js')(app, checkAdmin);
    require('./routes/plugins.js')(self, app, checkAdmin);
    require('./routes/logs.js')(self, app, checkAdmin, config);
    require('./routes/file_transfer.js')(self, app, checkAdmin, config);

    }
    catch(e){
        console.log("exception with require routes", e)
    }



    var dbc = redis.createClient(config.redis_port);
    dbc.on("error", function(err){
         console.log( 'database error:', err);
         return false;
    });

    var ACTIVE_SESSIONS = {};
    var get_active_sessions = function(callback){
        dbc.get('sess:active',
            function(err, sessionIDs){
                if( err || !sessionIDs ){
                    console.log( 'error:', err);
                    return false;
                }
                ACTIVE_SESSIONS = {};
                var activeSessions = sessionIDs.split(',');
                _.each(activeSessions, function(sess){
                    matches = sess.match(/sess:([0-9]+):([A-Z]+):([0-9]+).*/);
                    if( matches.length == 4 ){
                        var deviceID = matches[1];
                        var capabilities = matches[2];
                        if( !(ACTIVE_SESSIONS[deviceID]) ){ACTIVE_SESSIONS[deviceID]={}}
                        var sessData = {
                            device: matches[1]
                            ,capabilities: matches[2]
                            ,session_id: matches[3]
                            ,db_key: 'sess:'+matches[1]+':'+matches[2]
                        }
                        if( capabilities.search(/C/) != -1 ){
                            ACTIVE_SESSIONS[deviceID].command = sessData;
                        }
                        if( capabilities.search(/E/) != -1 ){
                            ACTIVE_SESSIONS[deviceID].video = sessData;
                        }
                        if( capabilities.search(/I/) != -1 ){
                            ACTIVE_SESSIONS[deviceID].ifb = sessData;
                        }

                    }
                });
                if( typeof(callback) == "function" ){
                    callback(ACTIVE_SESSIONS);
                }
        });

    };
    get_active_sessions();

    // Routes

    app.get('/', function(req, res){
        res.render('index', {});
    });

    app.get('/dashboard', function(req, res){
        res.render('dashboard', {});
    })


    app.get('/version', function(req, res){
        try{
            res.header("Content-Type", "text/plain");
            var filepath = '/opt/teradek/sputnik/VERSION';
            //monkey-patch of express bug
            res.sendfile(path.basename(filepath), {root: path.dirname(filepath)});
        }
        catch(e){
            console.log('failed to get Version file', fielpath, e);
        }
    });
    app.get('/session', function(req, res){
        res.render('session', {});
    });

    app.get('/session/:sessID', function(req, res){
        res.render('session', {
            sessID: req.params.sessID
        });
    });

    app.get('/session/:sessID/:connID', function(req, res){
        res.render('connection', {
            sessID: req.params.sessID
            ,connID: req.params.connID
            ,startTime: req.params.startTime
            ,endTime:   req.params.endTime

        });
    });

    app.get('/config/viewOnly', function(req,res){
        if( req.isAuthenticated() && req.user && req.user.username == 'guest' ){
            res.send('true');
            return true;
        }
        else{
            res.send('false');
            return false;
        }
    })

    app.get('/restart', checkAdmin, function(req, res){
        if( config.view_only ){
            res.writeHead(200, {"Content-Type": "application/json"});
            res.write('{"error":"You are not authorized to perform this action."}');
            res.end();
            return false;
        }
        res.send('attempting restart');
        exec("service sputnik restart", function(error, stdout, stderr){
                //res.writeHead(200, {"Content-Type": "application/json"});
                if (error) {
                    //res.write('{"error":"' + error + '"}');
                    console.log('failed to restart sputnik', error, stdout, stderr);
                }
                else{
                    //var output = stdout.replace(/\n/g, " ");
                    //res.write('{"success":"' + output + '"}');
                }
                //res.end();
        });
    });


    app.get('/connection', function(req, res){
        res.render('connection', {});
    });
    app.get('/dashboard', function(req, res){
        res.render('dashboard', {});
    });

    app.get('/json/:sessID/port/:portID', function(req, res){
        var sessID = req.params.sessID;
        var portID = req.params.portID;
        if( sessionArgsIsInvalid(sessID) ){
            res.write('{"error":"session is invalid."}');
            res.end();
            return false;
        }
        var db_key = ACTIVE_SESSIONS[sessID].video.db_key;

        res.writeHead(200, {"Content-Type": "application/json"});
        try{
            console.log( 'sessid', sessID, portID, 'sessid', 'portid')

            dbc.zrange(db_key+ ':conn:'+portID+':modem_info', 0, 0,
                     function(err, reply){
                        if( err ){ console.log( 'error:', err); return false;}
                        if (reply && reply[0] && typeof(reply[0]) == 'string') {
                            res.write(reply[0]);
                        }
                        res.end();
                      });
        }
        catch(e){
            res.write('{"error":"unable to create database client"}');res.end();
            return false;
        }
    });

    app.get('/json/sessions/totals', function(req, res){
        try{
            var resObj = {};
            get_active_sessions(function(sessionKeys){

                if( sessionKeys == null || Object.keys(sessionKeys).length === 0){
                    res.json({});
                    return false;
                }
                var resObj = {};
                var completeResponse = function(){
                    //make sure all sessions have an array (not null) signifying all db queries done
                    if( _.every(resObj, function(sess){return sess != null;}) ){
                        res.json(resObj);
                    }
                }
                _.each(sessionKeys, function(sess, sessionID){
                    resObj[sessionID] = null;
                });
                _.each(sessionKeys, function(sess, sessionID){
                    var video_session = sess.video;
                    if( video_session ){
                        dbc.zrange(video_session.db_key+':conn:total:stats', -20, -1, function(err, replies){
                            if( replies ){
                                //convert response with array of JSON string into valid JSON
                                var data = [];
                                _.each(replies, function(r){
                                    var jsonData = JSON.parse(r);
                                    if( jsonData.session_id == video_session.session_id ){
                                        data.push(jsonData);
                                    }
                                });
                                resObj[sessionID] = data;
                            }
                            else{
                                resObj[sessionID] = [];
                            }
                            completeResponse();
                        });
                    }
                    else{
                        resObj[sessionID] = {error:"Video data unavailable."};
                        completeResponse();
                    }

                })
            });
        }
        catch(e){
            console.log('unable to get active sessions', e);
            res.write('{"error":"unable to create database client"}');res.end();
            return false;
        }
    });

    app.get('/uptime/:sessID', function(req, res){
        res.writeHead(200, {"Content-Type": "application/json"});
        var sessionID = req.params.sessID;
        if( sessionArgsIsInvalid(sessID) ){
            res.write('{"error":"session is invalid."}');
            res.end();
            return false;
        }
        var db_key = ACTIVE_SESSIONS[sessID].video.db_key;
        dbc.zrange(db_key + ':conn:total:stats', 0, 0, function(err, reply){
            if (reply && reply[0] && typeof(reply[0]) == 'string') {
                res.write(reply[0]);
            }
            res.end();
        });


    });

    app.get('/time', function(req, res){
        res.writeHead(200, {"Content-Type": "text/plain"});
        var ts = new Date().getTime();
        res.write(ts.toString());
        res.end();
    });


    var sessionArgsIsInvalid = function(sessID){
        return ( !(sessID) ||
                !( ACTIVE_SESSIONS[sessID] ) ||
                !( ACTIVE_SESSIONS[sessID].video) ||
                !( ACTIVE_SESSIONS[sessID].video.db_key));
    }


    app.get('/:sid/snapshot', function(req, res){

       var sid = req.params.sid;
       if( !(self._sessions[sid]) ){
           res.json({error:'session does not exist'});
           return false;
       }
       //var mime = require('mime');
       //console.log("MIME", mime.lookup('ppm'));
       var sess = self.sess(sid);
       if( '_' in req.query ){delete req.query._}

       sess.iface('E').enqueueOne('dev:emit-snapshot', function (session) {
       session.takeSnapshot(req.query).then(function(data){
            var mimeTypes = {
               'ppm':'image/x-portable-pixmap'
               ,'h264':'image/h264'
               ,'jpeg':'image/jpeg'
            }
            var mimeType = mimeTypes['jpeg'];
            if( 'type' in req.query ){
                mimeType = mimeTypes[req.query.type];
            }
            res.writeHead(200, {"Content-Type": mimeType});
            res.write(data);
            res.end();

        }, function(){
             res.json({error:'Failed to generate snapshot'});
            });
       });
    });


    app.get('/:sid/restartCodec', function(req, res){

       var sid = req.params.sid;
       if( !(self._sessions[sid]) ){
           res.json({error:'session does not exist'});
           return false;
       }
       //var mime = require('mime');
       //console.log("MIME", mime.lookup('ppm'));
       var sess = self.sess(sid);

       sess.iface('C').enqueueOne('dev:restart-codec', function (session) {
       session.restartCodec().then(function(data){
            res.json(data);
        }, function(){
             res.json({error:'Failed to generate snapshot'});
            });
       });
    });




    app.get('/json/:sid/settings', function(req, res){
       var sid = req.params.sid;
       var q = req.query.q;
       var sess = self.sess(sid);

       sess.iface('C').enqueueOne('dev:emit-getSettings', function (session) {
       session.getSettingsIntrospect(q).then(function(data){
            res.json(data);
        }, function(data){
             res.json({error:'Failed to get settings', data:arguments});
            });
       });
    });


    app.post('/json/:sid/settings', function(req, res){
       var sid = req.params.sid;
       var sess = self.sess(sid);
       var settings = req.body;
       console.log('set settinga data', settings)
       sess.iface('C').enqueueOne('dev:emit-setSettings', function (session) {
       session.setSettings(settings).then(function(data){
            res.json(data);
        }, function(data){
             res.json({error:'Failed to set settings', data:arguments});
            });
       });
    });


    /*
     *  note: important to distinguish that 'connection' is not the same as 'port(ppp#)'
     *  as theoretically you may have multiple connections per port
     *  but as a shortcut connection is ignored
     */
    app.get('/json/:sessID', function(req, res){
        res.writeHead(200, {"Content-Type": "application/json"});
        var startTime = req.query['startTime'];
        var endTime = req.query['endTime'];
        if( !(endTime) ){
          endTime = "+inf";
        }
        var sessID = req.params.sessID;
        if( sessionArgsIsInvalid(sessID) ){
            res.write('{"error":"session is invalid."}');
            res.end();
            return false;
        }
        var db_key = ACTIVE_SESSIONS[sessID].video.db_key;
        dbc.keys(db_key+':conn:*', function(err, conns){
            if (!(conns) || !(conns.length)) {
                console.log('no connections found', err);
                if ( err ){
                    res.write('{"error":"'+err+'"}');
                }
                else{
                    res.write("{}");
                }
                res.end();
                return false;
            }
            var response = "{";
            connsLength = conns.length;
            if( err ){ console.log( 'error:', err); return false;}
            conns.forEach(function(connKey, connIndex){
                dbc.zrangebyscore(connKey, startTime, endTime, function(err, replies){
                    if(connKey.search('total') != -1){
                        var portName = 'total';
                    }
                    else{
                        var portName = "Unknown";
                       replies.every(function(replyStr){
                         try{
                         var firstReply = JSON.parse(replyStr);

                         if( firstReply.name && firstReply.name != "" ){
                              portName = firstReply.name;
                              return false;
                         }
                         return true;
                         }
                         catch(e){
                             console.log( 'database range port', e );

                         }
                       })
                    }
                    response += '"' + portName + '":';
                    response += '[ ';
                    replies.forEach(function(reply, i){
                        response += reply + ',';
                    });
                    response = response.substring(0, response.length - 1);
                    response += '],';
                    //last connection
                    if (connIndex == connsLength - 1) {
                        response = response.substring(0, response.length - 1);
                        response += '}';
                        res.write(response);
                        res.end();
                    }
                });
            });
        });
    });


    app.get('/client', function(req, res){
        res.render('client', {});
    });


    var socketArgsIsInvalid = function(data){
        return (typeof data != 'object' ||
                !(data.sessID) ||
                !( ACTIVE_SESSIONS[data.sessID] ) ||
                !( ACTIVE_SESSIONS[data.sessID].command) ||
                !( ACTIVE_SESSIONS[data.sessID].command.db_key));
    }

    var getRedisClient = function(sessID){
        if( ACTIVE_SESSIONS[sessID] ){
            var db_subscribe = ACTIVE_SESSIONS[sessID].command.db_key;
            if( !(REDIS_CLIENTS[ db_subscribe ]) ){
                REDIS_CLIENTS[ db_subscribe ] = {'pub':redis.createClient(), 'sub':redis.createClient()};
                REDIS_CLIENTS[ db_subscribe ]['sub'].subscribe(db_subscribe);
                REDIS_CLIENTS[ db_subscribe ]['sub'].on("message", function( channel, message ){
                    try{
                    if( message.toString().substr(0,5) == 'reply' ){
                        var startPayload = message.indexOf(':', 6)+1;
                        var payload = JSON.parse(message.substr(startPayload));
                        //if Link Mapping request
                        if( payload.cube_id ){
                            var payload = channel.split(':');
                            if( payload.length > 2 ){
                               payload.link_id = [1];
                               devicesSocket.emit('new_settings', payload);
                            }
                        }
                        else{
                            devicesSocket.to(sessID).emit('new_settings', payload);
                        }
                    }
                    else if( message == "close_session" ){
                        devicesSocket.to(sessID).emit('close_session');
                    }
                    else if( message == "open_session" ){
                        devicesSocket.to(sessID).emit('open_session');
                    }
                    }
                    catch(e){
                        //console.log('Error', e);
                    }
                })
            }
            return REDIS_CLIENTS[ db_subscribe ];
        }
        return redis.createClient();
    }


    devicesSocket.on('connection', function(socket){
        socket.on('leave', function(data){
            socket.leave(data.sessID);
        })

        socket.on('join', function(data, fn){
            if( socketArgsIsInvalid(data) ){
                fn({error:'Session does not exist or does not have remote control enabled.'});
                return false;
            }
             var sessID = data.sessID;
            socket.join(sessID);
            socket.broadcast.to(sessID).emit('joined', socket.id);
            getRedisClient(sessID);
            if( fn ){
                fn(socket.id);
            }

        })


         /*  ---------- Clients will request these ------------- */
        socket.on('client_get_settings', function(data){
            if( socketArgsIsInvalid(data) ){
                return false;
            }
            var sessID = data.sessID;
            var jsonCmd = data.jsonCmd;
            var db_publish = ACTIVE_SESSIONS[data.sessID].command.db_key;
            var rc = getRedisClient(sessID);
            rc['pub'].publish(db_publish,'req:124:2:1:3:'+jsonCmd);
        })

        socket.on('client_action', function(data){
            if( config.view_only ){
                socket.send("unauthorized");
                return false;
            }
            if( socketArgsIsInvalid(data) ){
                return false;
            }
            var sessID = data.sessID;
            var jsonCmd = data.jsonCmd;
            var ioctl = ( data.ioctl ) ? data.ioctl : 4;
            var db_publish = ACTIVE_SESSIONS[data.sessID].command.db_key;
            var rc = getRedisClient(sessID);
            rc['pub'].publish(db_publish,'req:124:2:'+ioctl+':3:'+jsonCmd);
        })


        socket.on('client_get_link_cube', function(data){
            var linkID = data.linkID;
            var jsonCmd = data.jsonCmd;
            var db_publish = ACTIVE_SESSIONS[linkID].command.db_key;
            var rc = getRedisClient( linkID );
            rc['pub'].publish(db_publish,'req:123:2:5:3:'+jsonCmd);
        })
    });

function requireAuthentication(req, res, next) {
      if( req.isAuthenticated() ){
          return next();
      }
      res.redirect('/login');
      return false;
}

function checkAdmin(req, res, next) {
      if( req.isAuthenticated() && req.user && req.user.username == 'admin' ){
          return next();
      }
      res.redirect('/login');
      return false;
}

    app.listen(config.web_port);
    console.log("Express server listening on port %d in %s mode", config.web_port, app.settings.env);

}


// in all events bind's this=app
var events = {
    'app:created': createWebServer
};



module.exports = function (app) {

    _.each(events, function (handler, evtName) {
        app.on(evtName, handler.bind(app));
    });

}
