var _ = require('underscore'),
    deferred = require('deferred'),
    Events = require('events').EventEmitter,
    rtmpSessCounter = 0;

function getSessId(sid) {
    rtmpSessCounter++;
    return sid + 'stream' + rtmpSessCounter;
}

module.exports = function (session, encSid, params) {

    console.log('streamToCDN()...');

    /*
     To start:
     ioctl: 0x2003
     payload: "ip=1.15940683.fme.ustream.tv,session=25501337stream1,port=1935,file=ustreamVideo/15940683,
     user_agent=FMLE,copyright=Teradek,device=cube2,compat=0,strm=dNwWKPNSgycNvK6cAJbGwXVX2ugrscVV,username=,password="

     To disconnect:
     ioctl: 0x2004
     same payload

     State values from rtmpwriter().getState():
     0 : disconnected
     1 : connected
     2 : connecting
     3 : reconnecting

     To test, set up encoder with:
     RTMP Url: rtmp://1.15940683.fme.ustream.tv/ustreamVideo/15940683
     RTMP Stream: dNwWKPNSgycNvK6cAJbGwXVX2ugrscVV


     This plugin requires that initial parameters are passed to the plugin to initialize it
     and then another set of parameters to specify which CDN to stream the RTMP stream to



     */

    var self = this,
        def = deferred(),
        events = new Events,
        ses = getSessId(encSid),//params ||
        init_params =  [
            'prt=0',
            'tprt=0',
            'ses=' + ses,
            'nif=ANY',
            'abt=96000',
            'vbt=4000000',
            'fr1001=1',
            'fr=30',
            'tick_120=%!120KHZ_TICKS',
            'use_sender=0',
            'stoc=4096',
            'avf=2'
        ];
        if( params ){
            _.map(params, function(el){
                init_params.push('pub.'+el);
            })
        }

    events.setMaxListeners(0);

    /* --- Connect plugin --- */

    events.on('rtmp:connect', function (pluginId, params) {
        session.rtmpwriter(ses).connect(pluginId, params)(function () {
            setTimeout(function () {
                events.emit('rtmp:get-state', pluginId, params);
            }, 1000);
        }, function (err) {
            events.emit('rtmp:connect-fail', err);
        });
    });

    /* --- Get state of plugin --- */

    events.on('rtmp:get-state', function (pluginId, params) {
        session.rtmpwriter(ses).getState(pluginId)(function (data) {

            console.log('rtmp:get-state result:', data);

            if (data == 0) {
                console.log('attempting to re-connect RTMP connection', pluginId);
                events.emit('rtmp:connect', pluginId, params);
            } else if (data == 1) {
                events.emit('rtmp:connected');
            }
        }, function (err) {
            events.emit('rtmp:connect-fail', err);
        });
    });

    /* --- Plugin connect success --- */

    events.on('rtmp:connected', function () {
        console.log('streamToCDN() success! - NOW WHAT?');
        events.removeAllListeners();
    });

    /* --- Plugin connect fail --- */

    events.on('rtmp:connect-fail', function (err) {
        console.log('streamToCDN() failed:', err.toString());
        events.removeAllListeners();
    });

    /* --- Add and load plugin --- */
    session.addPluginParams('rtmp', init_params)(function (pluginId) {
        session.loadPlugin(pluginId)(function () {
            events.emit('rtmp:connect', pluginId, params);
            def.resolve(pluginId);
        }, function (err) {
            events.emit('rtmp:connect-fail', err);
            def.resolve(err);
        });
    }, function (err) {
       events.emit('rtmp:connect-fail', err);
        def.resolve(err);
    });

    return def.promise;
}