var properties = require('properties')
            ,_ = require('underscore')
           ,fs = require('fs')
    ,validator = require('validator')
    ,path      = require('path')
     ,deferred = require('deferred');

/*_.extend(validation.validators, {
  portnumber: function(value, attr, customValue, model) {
      console.log("IT BE RUNNING YO!", value, attr ,validation.validators.range)
    return validaton.validators.range(value, attr, [1,50009], model);
  }
  ,string_set: function(value, attr, customValue, model) {
    return this.oneOf(value, attr, customValue, model);
  }
});*/

validator.extend( 'is_integer_range', function( input, min, max ){
    var input = validator.toInt(input);
    return input >= min && input <= max;
});

validator.extend( 'is_portnumber', function( input ){
    return validator.is_integer_range(input, 1, 50009);
});

validator.extend( 'is_string_set', function( input, values ){
    return validator.isIn(input,values);
});

validator.extend( 'is_string', function( input ){
    return validator.isAscii(input);
});

validator.extend( 'is_blank', function( input ){
    return true;
});

validator.extend( 'is_base64', function( input ){
    return validator.matches(input, /^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{4}|[A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)$/);
});

validator.extend( 'is_write_path', function( input ){
    return /[^\0]+/.test(input);
    //remove checking if path actually exist since sputnik creates the folder path on start
    try{
        var stats = fs.statSync(input);
        if( !(stats.isDirectory()) ){
            return false;
        }
        do{
            var testFile = path.join(input, Math.random().toString());
        }
        while( fs.existsSync(testFile) );

    var fd = fs.openSync(testFile, 'a');
    fs.closeSync(fd);
    fs.unlinkSync(testFile);

    return true;
    }
    catch(e){
        console.log(e);
        return false;
    }

});




var options = {
  path: true
};

var configFile = "/etc/sputnik.conf";
var configMeta = {
    input_port:{
         label:"Video input port"
        ,description:"The input port that Sputnik listens on for incoming connections from"
        ,default:5111
        ,type:['portnumber']
    }
    ,web_port:{
         label:"Webserver port"
        ,description:"The port that Sputnik dashboard is presented on"
        ,default:1957
        ,type:['portnumber']
    }
    ,redis_port:{
         label:"Database port"
        ,description:"The port that Sputnik Redis database listens on"
        ,default:6379
        ,type:['portnumber']
    }
    ,log_size:{
         label:"Server Log file size"
        ,description:"Maximum size of server log file before rotation"
        ,default:10485760
        ,type:['integer_range']
        ,max:1048576000
        ,min:500
    }
    ,log_number:{
         label:"Server Log file size"
        ,description:"Maximum number of server log files before deleting oldest"
        ,default:5
        ,type:['integer_range']
        ,max:500
        ,min:1
    }
    ,core_client:{
         label:"Core client enabled"
        ,description:"Enable Core uplink"
        ,default:'no'
        ,choices:['yes','no']
        ,type:['string_set']
    }
    ,rtp_server:{
         label:"Enable RTP Server"
        ,description:"# A local RTP Server for monitoring incoming video streams"
        ,default:"no"
        ,choices:['yes','no']
        ,type:['string_set']
    }
    ,rtp_server_port:{
         label:"RTP Server Port"
        ,description:"The port that the RTP Server listens on"
        ,default:554
        ,type:['portnumber']
    }
    ,rtp_server_username:{
         label:"RTP Server username"
        ,description:"The username used to authenticate access to the RTP Server streams. Leave empty for no authentication."
        ,default:''
        ,type:['blank','string']
    }
    ,rtp_server_password:{
         label:"RTP Server password"
        ,description:"The password used to authenticate access to the RTP Server streams. Leave empty for no authentication."
        ,default:''
        ,type:['blank','base64']
    }
    ,view_only:{
        label:"View Only"
        ,description:"will probably be deprecated"
        ,choices:['yes','no']
        ,type:['string_set']
    }
    ,upload_path:{
        label:"Upload Path"
        ,description:"path to upload files"
        ,type:['write_path']
    }
}

module.exports = {
        settings:null
        ,load: function( def ){
          var _this = this;
          properties.parse( configFile, options,
              function( error, settings ){
                if( error ){
                    console.log('Error reading config file:', configFile, error);
                    return def.resolve(error);
                }
                _this.settings = settings;
                def.resolve(settings);
          });
        }
        ,get: function(){
            var def = deferred();
            this.load( def );
            return def.promise;
        }
        ,geti:function(){
            var def = deferred();
            var _this = this;
            this.get()(function(settings){
                var ret = _.map(configMeta, function(v,k){
                    var newObj = _.extend({id:k}, v);
                    if( settings[k] ){
                        _.extend(newObj, {value:settings[k]});
                    }
                    return newObj;
                });
                console.log(_this.validate(ret))
                def.resolve(ret);
            })
            return def.promise;
        }
        ,validate:function( settings ){
            var error_status = {};
            var mergeSettings = _.map(settings, function(v, k){
                if( configMeta[k] ){
                    var obj = configMeta[k];
                    obj.value = v;
                    return obj;
                }
                else{
                    console.log('ERROR: no meta definition for ', k);
                }
            });

            _.each(mergeSettings, function( v ){
                var args = [];
                _.every(v.type, function(type){
                    args.push(v.value);
                    if( 'choices' in v ){
                         args.push(v.choices);
                    }
                    if( 'min' in v ){
                        args.push(v.min);
                    }
                    if( 'max' in v ){
                        args.push(v.max);
                    }

                    if( validator.trim(v.value) == '' && type == 'blank' ){
                        return false;
                    }
                    if( !(validator['is_'+type].apply(null, args)) ){
                        error_status.error = true;
                        if( !(error_status.errorMsgs)){ error_status.errorMsgs = []}
                        error_status.errorMsgs.push({name:v.label, msg:"invalid value for "+type});
                    }
                })
            });
            return error_status;
        }
        ,set: function( settings ){
            console.log("trying to set these sputnik settings", settings);
            var def = deferred();

            var errors = this.validate(settings);
            console.log('errors validation failure', errors)
            if( 'error' in errors ){
                def.reject(new Error(JSON.stringify(errors)));
                return def.promise;
            }

            fs.readFile(configFile, 'utf8', function (err,data) {
              if (err) {
                console.log(err);
                def.resolve(err);
              }
              var result = data;
              _.each(settings, function(v, k){
                  var regex = new RegExp("^"+k+"\\s*=.*$", 'gm');
                  if( result.match(regex) ){
                    result = result.replace( regex, k+'='+v );
                  }
                  else{
                      settings[k] = '##Failed to set this value: it does not exist in conf file #';
                  }

              })
              fs.writeFile(configFile, result, 'utf8', function (err) {
                 if (err){
                    console.log(err);
                    def.resolve(err);
                 }
                 def.resolve(settings);
              });
            });
            return def.promise;
        }
    }